// $Id: DownloadItem.cs 20 2008-03-26 02:51:32Z arcao $
//
// Copyright (C) 2008 Martin "Arcao" Sloup <arcao@arcao.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using OSD.PluginBase;

namespace OSD.Program
{

    #region Enum DownloadItemStatus

    ///<summary>
    /// Vycet stavu, ktere muze mit polozka
    ///</summary>
    public enum DownloadItemStatus
    {
		// Nasledujici stavy se ukladaji spolecne s polozkou v nastaveni programu
    
		/// <summary>
		/// Polozka ceka na stahnuti
		/// </summary>
        Queued = 1,
        /// <summary>
        /// Polozka je stazena
        /// </summary>
        Downloaded = 2,
        /// <summary>
        /// Polozka je smazana na vzdalenem serveru
        /// </summary>
        Deleted = 3,
        /// <summary>
        /// Polozka je pozastavena
        /// </summary>
        Paused = 4,
        
        //-----------

		// Nasledujici stavy se ukladaji spolecne s polozkou v nastaveni programu jako stav Queued
		/// <summary>
		/// Polozka se prave stahuje
		/// </summary>
        Downloading = 5,
        /// <summary>
        /// Polozka ceka po neuspesnem stahovani na dalsi pokus
        /// </summary>
        Waiting = 6
    }

    #endregion

    #region DownloadItemStatusChangedEventArgs

	/// <summary>
	/// Trida zastresujici hodnotu argumentu u udalosti DownloadItemStatusChanged
	/// </summary>
    public class DownloadItemStatusChangedEventArgs : EventArgs
    {
        private readonly DownloadItemStatus status;
        private readonly string statusText;

		/// <summary>
		/// Konstruktor tridy
		/// </summary>
		/// <param name="status">stav polozky</param>
		/// <param name="statusText">textovy popis stavu polozky</param>
        public DownloadItemStatusChangedEventArgs(DownloadItemStatus status, string statusText)
        {
            this.status = status;
            this.statusText = statusText;
        }

		/// <summary>
		/// Vraci stav polozky
		/// </summary>
        public DownloadItemStatus Status
        {
            get { return status; }
        }

		/// <summary>
		/// Vraci textovou reprezentaci stavu polozky
		/// </summary>
        public string StatusText
        {
            get { return statusText; }
        }
    }

	/// <summary>
	/// Trida zastresujici hodnotu argumentu u udalosti DownloadItemStatusChanged s urcenim casu pro cekani
	/// </summary>
    public class TimeoutDownloadItemStatusChangedEventArgs : DownloadItemStatusChangedEventArgs
    {
        private readonly int suggestedWaiting = 0;

		/// <summary>
		/// Konstruktor tridy
		/// </summary>
		/// <param name="status">stav polozky</param>
		/// <param name="statusText">textovy popis stavu polozky</param>
		/// <param name="suggestedWaiting">doporuceny cas cekani</param>
        public TimeoutDownloadItemStatusChangedEventArgs(DownloadItemStatus status, string statusText,
                                                         int suggestedWaiting)
            : base(status, statusText)
        {
            this.suggestedWaiting = suggestedWaiting;
        }

		/// <summary>
		/// Vraci doporuceny cas cekani
		/// </summary>
        public int SuggestedWaiting
        {
            get { return suggestedWaiting; }
        }
    }

    #endregion

    ///<summary>
    /// Trida zastresujici vsechny informace o polozce
    ///</summary>
    public class DownloadItem : IDisposable
    {
        private string url;
        private string description = "";
        private Download download = null;
        private string newFilename = "";
        private string password = "";
        private DownloadItemStatus status;
        private string statusText = "";
        private int timeToGo = 0;

		/// <summary>
		/// Konstruktor tridy s nastaveni url polozky a jeho pocatecniho stavu
		/// </summary>
		/// <param name="url">url polozky</param>
		/// <param name="status">pocatecni stav polozky</param>
        public DownloadItem(string url, DownloadItemStatus status)
        {
            this.url = url;
            this.status = status;
        }

		/// <summary>
		/// Konstruktor tridy s nastavenim url polozky
		/// </summary>
		/// <param name="url">url polozky</param>
        public DownloadItem(string url) : this(url, DownloadItemStatus.Queued) {}

        #region Getters/Setters

		/// <summary>
		/// Vraci/Nastavuje url polozky. Pokud se pri nastveni url polozka jiz stahuje, bude zastaven proces stahovani a znovu psusten s novou url.
		/// </summary>
        public string Url
        {
            get { return url; }
            set
            {

                if (url == value) return;

                //pokud nastala zmena url restartujeme stahovani
                url = value;
                if (status != DownloadItemStatus.Paused || status != DownloadItemStatus.Downloading)
                    ResetStatus();
                else if (status == DownloadItemStatus.Downloading)
                    doDownload();
            }
        }

		/// <summary>
		/// Vraci/Nastavuje heslo polozky
		/// </summary>
        public string Password
        {
            get { return password; }
            set { password = value; }
        }

		/// <summary>
		/// Vraci/Nastavuje novy nazev souboru, kam se bude polozka ukladat
		/// </summary>
        public string NewFilename
        {
            get { return newFilename; }
            set { newFilename = value; }
        }

		/// <summary>
		/// Vraci nastavuje popis polozky
		/// </summary>
        public string Description
        {
            get { return description; }
            set { description = value; }
        }

		/// <summary>
		/// Vraci Download tridu, ktera se stara o stahnuti polozky, pokud bezi proces stahovani, jinak vraci null
		/// </summary>
        public Download Download
        {
            get { return download; }
            //set { download = value; }
        }

		/// <summary>
		/// Vraci true, pokud je polozka oznacena jako smazana na strane serveru
		/// </summary>
        public bool Deleted
        {
            get { return (status == DownloadItemStatus.Deleted); }
        }

		/// <summary>
		/// Vraci/nastavuje, zda polozka je pozastavena (true), ci neni (false)
		/// </summary>
        public bool Paused
        {
            get { return (status == DownloadItemStatus.Paused); }
            set
            {
                if (download != null)
                {
                    // uklidime
                    download.Dispose();
                    download = null;
                }
                status = value ? DownloadItemStatus.Paused : DownloadItemStatus.Queued;
                if (StatusChanged != null)
                    StatusChanged(this, new DownloadItemStatusChangedEventArgs(status, string.Empty));
            }
        }

		/// <summary>
		/// Vraci true, pokud je polozka jiz stazena
		/// </summary>
        public bool Downloaded
        {
            get { return (status == DownloadItemStatus.Downloaded); }
        }

		/// <summary>
		/// Vraci stav polozky
		/// </summary>
        public DownloadItemStatus Status
        {
            get { return status; }
        }

		/// <summary>
		/// Vraci textovou reprezentaci polozky. Pokud ma polozka TimeToGo > 0 tak dale obsahuje "Cekam x s...", kde x je hodnota TimeToGo.
		/// </summary>
        public string StatusText
        {
            get
            {
                if (timeToGo > 0)
                {
                    if (statusText.Length > 0)
                        return string.Format("{0} ekm {1} s...", statusText, timeToGo);
                    else
                        return string.Format("ekm {0} s...", timeToGo);
                }
                else
                {
                    return statusText;
                }
            }
        }

		/// <summary>
		/// Vraci/Nastavuje hodnotu cekani na dalsi akci. Nastavovani by se melo pouzivat jen pri zmenseni cekani. Pro nastaveni pocatecniho cekani se musi pouzit metoda Wait.
		/// </summary>
        public int TimeToGo
        {
            get { return timeToGo; }
            set { timeToGo = value; }
        }

		/// <summary>
		/// Nastavi stav polozky na Queued a hodnotu TimeToGo na 0
		/// </summary>
        public void ResetStatus()
        {
            timeToGo = 0;
            status = DownloadItemStatus.Queued;
			statusText = "";
			if (StatusChanged != null)
				StatusChanged(this, new DownloadItemStatusChangedEventArgs(status, statusText));
        }

		/// <summary>
		/// Nastavuje cekani na daalsi akci.
		/// </summary>
		/// <param name="delay">delka cekani</param>
		/// <param name="timeoutWaiting">jedna se o cekani, pri kterem bude snovu spusten proces stahovani (true), nebo se jedna o cekani na time-slot (false)</param>
        public void Wait(int delay, bool timeoutWaiting)
        {
            timeToGo = delay;
            if (timeoutWaiting)
                status = DownloadItemStatus.Waiting;
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Uzavre vsechny pouzite prostredky, ukonci vsechny stahovani.
        /// </summary>
        public void Dispose()
        {
            DestroyDownload();
        }

        #endregion

		/// <summary>
		/// Handler pro udalost, ktera nastane pri zmene stavu polozky (napriklad pri stahovani) 
		/// </summary>
        public event EventHandler StatusChanged;
		/// <summary>
		/// Handler pro udalost, ktera nastane pri zmene stavu stahovani v Download tride
		/// </summary>
        public event EventHandler DownloadStatusChanged;
		/// <summary>
		/// Handler pro udalost, ktera nastane pri zmene poctu stahnutych dat 
		/// </summary>
        public event EventHandler PositionChanged;
		/// <summary>
		/// Handler pro udalost, ktera nastane pri dokonceni stahovani
		/// </summary>
        public event EventHandler DownloadComplete;
		/// <summary>
		/// Handler pro udalost, ktera nastane pri neuspesnem stahnuti
		/// </summary>
        public event EventHandler DownloadFailed;

        /// <summary>
        /// Spusti stahovani polozky
        /// </summary>
        /// <returns>reference na tento objekt</returns>
        public DownloadItem doDownload()
        {
            Plugins p = Plugins.GetInstance();
            foreach (DownloadPlugin plugin in p)
            {
                if (plugin.IsSuitableUrl(url))
                {
                    //uklidime
                    DestroyDownload();

                    ResetStatus();

                    download = plugin.GetDownloadForUrl(url);
                    download.DownloadFailed += OnDowloadFailed;
                    download.DownloadComplete += OnDownloadComplete;
                    download.PositionChanged += OnPositionChanged;
                    download.StatusChanged += OnDownloadStatusChanged;
                }
            }

            return this;
        }

        /// <summary>
        /// Zrusi predchozi download, pokud bezi a odregistruje vsechny pouzite eventy
        /// </summary>
        public void DestroyDownload()
        {
            if (download != null)
            {
                download.DownloadFailed -= OnDowloadFailed;
                download.DownloadComplete -= OnDownloadComplete;
                download.PositionChanged -= OnPositionChanged;
                download.StatusChanged -= OnDownloadStatusChanged;

                download.Dispose();
                download = null;
            }
            //if (status != DownloadItemStatus.Downloaded || ) status = DownloadItemStatus.Queued;
        }

        private void OnDowloadFailed(object sender, EventArgs e)
        {
            DownloadFailedEventArgs dfea = (DownloadFailedEventArgs) e;

            status = dfea.Status == DownloadStatus.Deleted ? DownloadItemStatus.Deleted : DownloadItemStatus.Waiting;
            statusText = dfea.StatusText;

            if (StatusChanged != null) StatusChanged(this, new DownloadItemStatusChangedEventArgs(status, StatusText));
            if (DownloadFailed != null) DownloadFailed(this, e);

            // uklidime
            DestroyDownload();
        }

        private void OnDownloadComplete(object sender, EventArgs e)
        {
            status = DownloadItemStatus.Downloaded;
            statusText = string.Empty;

            if (StatusChanged != null) StatusChanged(this, new DownloadItemStatusChangedEventArgs(status, StatusText));
            if (DownloadComplete != null) DownloadComplete(this, e);

            // uklidime
            DestroyDownload();
        }

        private void OnPositionChanged(object sender, EventArgs e)
        {
            // predame dal
            if (PositionChanged != null) PositionChanged(this, e);
        }

        private void OnDownloadStatusChanged(object sender, EventArgs e)
        {
            //predat dal
            if (DownloadStatusChanged != null) DownloadStatusChanged(this, e);

            DownloadStatusChangedEventArgs dscea = (DownloadStatusChangedEventArgs) e;
            switch (dscea.Status)
            {
                case DownloadStatus.Connecting:
                case DownloadStatus.Downloading:
                case DownloadStatus.Waiting:
                    status = DownloadItemStatus.Downloading;
                    statusText = dscea.StatusText;
                    break;
                default:
                    //jiny stav ohlasuje prislusny OnXXXX
                    return;
            }

            //predat dal
            if (StatusChanged != null)
                StatusChanged(this, new DownloadItemStatusChangedEventArgs(status, statusText));
        }

        ///<summary>
        ///Determines whether the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>.
        ///</summary>
        ///
        ///<returns>
        ///true if the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>; otherwise, false.
        ///</returns>
        ///
        ///<param name="obj">The <see cref="T:System.Object"></see> to compare with the current <see cref="T:System.Object"></see>. </param><filterpriority>2</filterpriority>
        public override bool Equals(object obj)
        {
            if (obj is DownloadItem)
                if (Url.Equals(((DownloadItem) obj).Url)) return true;
            return false;
        }

        ///<summary>
        ///Serves as a hash function for a particular type. <see cref="M:System.Object.GetHashCode"></see> is suitable for use in hashing algorithms and data structures like a hash table.
        ///</summary>
        ///
        ///<returns>
        ///A hash code for the current <see cref="T:System.Object"></see>.
        ///</returns>
        ///<filterpriority>2</filterpriority>
        public override int GetHashCode()
        {
            return Url.GetHashCode();
        }
    }
}