﻿// $Id: WebDownloader.cs 21 2008-03-27 02:38:24Z arcao $
//
// Copyright (C) 2008 Martin "Arcao" Sloup <arcao@arcao.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.IO;
using System.Net;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Web;

namespace OSD.PluginBase
{
    /// <summary>
    /// determines what type of post to perform.
    /// </summary>
    public enum MethodType
    {
        /// <summary>
        /// Does a get against the source.
        /// </summary>
        Get,
        /// <summary>
        /// Does a post against the source.
        /// </summary>
        Post
    }

    /// <summary>
    /// Submits post data to a url.
    /// </summary>
    public class WebDownloader : IDisposable
    {
        private const string DEFAULT_CHARSET = "ISO-8859-1";
        /// <summary>
        /// Default size of received bufer for SaveToFile method.
        /// </summary>
        public const int READ_BUFFER_LENGTH = 65536;

        // HttpWebRequest Setters/Getters
        private string accept;
        private long addRange = 0;
        private string charset;
        private string connection;
        private string contentType;
        private CookieContainer cookieContainer;
        //private DateTime ifModifiedSince = null;

        private WebHeaderCollection headers;
        private MethodType method = MethodType.Get;
        private SortedDictionary<string, string> queryString = new SortedDictionary<string, string>();

        private int readBufferLength = READ_BUFFER_LENGTH;
        private string referer;
        private HttpWebResponse response = null;
        private string url = string.Empty;
        private string userAgent;

        /// <summary>
        /// Default constructor.
        /// </summary>
        public WebDownloader()
        {
            charset = DEFAULT_CHARSET;
            cookieContainer = new CookieContainer();
            accept = null;
            connection = null;
            contentType = null;
            referer = null;
            userAgent = null;
            headers = null;
        }

        /// <summary>
        /// Constructor that accepts a url as a parameter
        /// </summary>
        /// <param name="url">The url where the post will be submitted to.</param>
        public WebDownloader(string url)
            : this()
        {
            this.url = url;
        }

        /// <summary>
        /// Constructor allowing the setting of the url and items to post.
        /// </summary>
        /// <param name="url">the url for the post.</param>
        /// <param name="queryString">The values for the post.</param>
        public WebDownloader(string url, SortedDictionary<string, string> queryString)
            : this(url)
        {
            this.queryString = queryString;
        }

        /// <summary>
        /// Gets or sets the url to submit the post to.
        /// </summary>
        public string Url
        {
            get { return url; }
            set { url = value; }
        }

        /// <summary>
        /// Gets or sets the name value collection of items to post.
        /// </summary>
        public SortedDictionary<string, string> QueryString
        {
            get { return queryString; }
            set { queryString = value; }
        }

        /// <summary>
        /// Gets or sets the type of action to perform against the url.
        /// </summary>
        public MethodType Method
        {
            get { return method; }
            set { method = value; }
        }

        ///<summary>
        /// Return response stream of current sended request
        ///</summary>
        ///<exception cref="ResponseStreamNotAvaibleException">if response is not available</exception>
        public Stream ResponseStream
        {
            get
            {
                if (response == null) throw new ResponseStreamNotAvaibleException("Call Send method before.");
                return response.GetResponseStream();
            }
        }

        ///<summary>
        /// Return response text of current sended request
        ///</summary>
        public string ResponseText
        {
            get
            {
                using (Stream responseStream = ResponseStream)
                {
                    using (StreamReader readStream = new StreamReader(responseStream, Encoding.GetEncoding(charset)))
                    {
                        return readStream.ReadToEnd();
                    }
                }
            }
        }

        /// <summary>
        /// Get Response of current sended request
        /// </summary>
        public HttpWebResponse Response
        {
            get { return response; }
        }

        /// <summary>
        /// Getter/Setter for Cookie Container
        /// </summary>
        public CookieContainer CookieContainer
        {
            get { return cookieContainer; }
            set { cookieContainer = value; }
        }

        ///<summary>
        /// Gets or sets the value of the Accept HTTP header.
        ///</summary>
        public string Accept
        {
            get { return accept; }
            set { accept = value; }
        }

        /// <summary>
        /// Gets or sets the value of the Connection HTTP header.
        /// </summary>
        public string Connection
        {
            get { return connection; }
            set { connection = value; }
        }

        /// <summary>
        /// Gets or sets the value of the Content-type HTTP header.
        /// </summary>
        public string ContentType
        {
            get { return contentType; }
            set { contentType = value; }
        }

        /// <summary>
        /// Gets or sets the value of the Referer HTTP header.
        /// </summary>
        public string Referer
        {
            get { return referer; }
            set { referer = value; }
        }

        /// <summary>
        /// Gets or sets the value of the User-agent HTTP header.
        /// </summary>
        public string UserAgent
        {
            get { return userAgent; }
            set { userAgent = value; }
        }

        /// <summary>
        /// Specifies a collection of the name/value pairs that make up the HTTP headers.
        /// </summary>
        public WebHeaderCollection Headers
        {
            get { return headers ?? (headers = new WebHeaderCollection()); }
            set { headers = value; }
        }

        /// <summary>
        /// Gets or sets the max size of SaveToFile receive buffer.
        /// </summary>
        public int ReadBufferLength
        {
            get { return readBufferLength; }
            set { readBufferLength = value; }
        }

        /// <summary>
        /// Adds a byte range header to a request for a specific range from the beginning or end of the requested data.
        /// </summary>
        public long AddRange
        {
            get { return addRange; }
            set { addRange = value; }
        }

        #region IDisposable Members
        /// <summary>
        /// Dispose previous response
        /// </summary>
        public void Dispose()
        {
            if (response != null)
            {
                response.Close();
                response = null;
            }
        }

        #endregion

        /// <summary>
        /// Posts the supplied data to specified url.
        /// </summary>
        /// <returns>a string containing the result of the post.</returns>
        public void Send()
        {
            Dispose();

            HttpWebRequest request;

            // generate query parameters
            StringBuilder parameters = new StringBuilder();
            foreach (KeyValuePair<string, string> item in queryString)
            {
                if (parameters.Length > 0) parameters.Append("&");
                parameters.Append(item.Key);
                parameters.Append("=");
                parameters.Append(HttpUtility.UrlEncode(item.Value));
            }
            string postData = parameters.ToString();

            // create web request
            if (method == MethodType.Post)
                request = (HttpWebRequest) WebRequest.Create(url);
            else
                request = (HttpWebRequest) WebRequest.Create(url + (postData.Length > 0 ? "?" + postData : ""));

            // set user-defined headers
            if (accept != null) request.Accept = accept;
            if (connection != null) request.Connection = connection;
            if (contentType != null) request.ContentType = contentType;
            if (referer != null) request.Referer = referer;
            if (userAgent != null) request.UserAgent = userAgent;
            if (addRange > 0)
            {
                //request.AddRange((int)addRange);
                //HACK for very big files (>2GB), url: http://forums.microsoft.com/MSDN/ShowPost.aspx?PostID=499572&SiteID=1
                //call private bool AddRange(string rangeSpecifier, string from, string to)
                typeof (HttpWebRequest).GetMethod("AddRange", BindingFlags.Instance | BindingFlags.NonPublic).Invoke(
                    request, new object[] {"bytes", addRange.ToString(NumberFormatInfo.InvariantInfo), ""});
            }


            if (method == MethodType.Post)
            {
                request.Method = "POST";
                request.CookieContainer = cookieContainer;
                request.ContentType = "application/x-www-form-urlencoded";
                request.ContentLength = postData.Length;
                using (Stream writeStream = request.GetRequestStream())
                {
                    UTF8Encoding encoding = new UTF8Encoding();
                    byte[] bytes = encoding.GetBytes(postData);
                    writeStream.Write(bytes, 0, bytes.Length);
                }
            }
            else
            {
                request.Method = "GET";
                request.CookieContainer = cookieContainer;
            }

            try
            {
                response = (HttpWebResponse) request.GetResponse();
            } catch (WebException ex)
            {
				// try catch http error 416 Requested Range Not Satisfiable
                if (ex.Response != null) response = (HttpWebResponse) ex.Response;
				if (response == null || response.StatusCode != HttpStatusCode.RequestedRangeNotSatisfiable) throw;
            }
            charset = response.CharacterSet;
            if (charset.Length == 0) charset = DEFAULT_CHARSET;
        }

        /// <summary>
        /// Posts the supplied data to specified url.
        /// </summary>
        /// <param name="queryString">The values to post.</param>
        /// <returns>a string containing the result of the post.</returns>
        public void Send(SortedDictionary<string, string> queryString)
        {
            this.queryString = queryString;
            Send();
        }

        /// <summary>
        /// Perform save response stream to file. If is used AddRange and server response 206 Partial Content, recived bytes will be appended into existing file.
        /// </summary>
        /// <param name="filename">filename where te received bytes would be saved</param>
        /// <param name="handler">handler for progress change event</param>
		public void SaveToFile(string filename, SaveToFileProgressChange handler)
        {
            if (response == null) return;
            if (response.StatusCode == HttpStatusCode.RequestedRangeNotSatisfiable)
            {
                // probably file is complete
                return;
            }

            //IAsyncResult result = null;

			//create async operation and asyncHandler
			AsyncOperation asyncOp = AsyncOperationManager.CreateOperation(response);
        	SendOrPostCallback asyncHandler = delegate(object state)
        	{
        		if (handler != null)
        			handler(this, (SaveToFileProgressChangedEventArgs) state);
        	};

			#region Old Code
			/*using (Stream s = ResponseStream)
            {
                FileMode mode = FileMode.Create;
                if (addRange > 0 && Response.StatusCode == HttpStatusCode.PartialContent)
                    mode = FileMode.Append;

                using (FileStream fs = new FileStream(filename, mode, FileAccess.Write, FileShare.Read, readBufferLength))
                {
                    //prepare received buffer
                    byte[] buffer = new byte[readBufferLength];
                    int count;
                    long totalCount = 0;
                    long length = response.ContentLength;

                    // if is mode append, add the addRange position
                    if (mode == FileMode.Append)
                    {
                        totalCount += addRange;
                        length += addRange;
                    }

                    while ((count = s.Read(buffer, 0, readBufferLength)) > 0)
                    {
                        fs.Write(buffer, 0, count);
                        totalCount += count;
                        // if we have handler, run him in his own thread
                        //if (handler != null) result = handler.BeginInvoke(this, new SaveToFileProgressChangedEventArgs(length, totalCount), null, null);
						asyncOp.Post(asyncHandler, new SaveToFileProgressChangedEventArgs(length, totalCount));
                    }
                    //we wait on end of last handler
                    //if (handler != null) handler.EndInvoke(result);
                }
            }*/
			
			#endregion

			// create ManualResetEvent
        	ManualResetEvent allDone = new ManualResetEvent(false);

			using (Stream s = ResponseStream)
			{
				FileMode mode = FileMode.Create;
				if (addRange > 0 && Response.StatusCode == HttpStatusCode.PartialContent)
					mode = FileMode.Append;

				using (FileStream fs = new FileStream(filename, mode, FileAccess.Write, FileShare.Read, readBufferLength))
				{
					//prepare received buffer
					byte[] buffer = new byte[readBufferLength];
					long totalReceivedBytes = 0;
					long length = response.ContentLength;

					// if is mode append, add the addRange position
					if (mode == FileMode.Append)
					{
						totalReceivedBytes += addRange;
						length += addRange;
					}

					//asynchronous reading from stream
					AsyncCallback beginReadCallback = null;
					beginReadCallback = delegate(IAsyncResult ar)
					{
						try
						{
							int count = s.EndRead(ar);

							if (count == 0)
							{
								// download complete
								allDone.Set();
								return;
							}

							fs.Write(buffer, 0, count);

							totalReceivedBytes += count;
							asyncOp.Post(asyncHandler, new SaveToFileProgressChangedEventArgs(length, totalReceivedBytes));

							s.BeginRead(buffer, 0, readBufferLength, beginReadCallback, s);
						} catch
						{
							allDone.Set();
						}
					};
					

					s.BeginRead(buffer, 0, readBufferLength, beginReadCallback, s);

					//we wait on end of download
					allDone.WaitOne();
										
					//if (handler != null) handler.EndInvoke(result);
				}
			}

        }

		/// <summary>
        /// Perform save response stream to file. If is used AddRange and server response 206 Partial Content, recived bytes will be appended into existing file.
        /// </summary>
        /// <param name="filename">filename where te received bytes would be saved</param>
        public void SaveToFile(string filename)
        {
            SaveToFile(filename, null);
        }

        /// <summary>
        /// Clear the current temporary request headers and request query string 
        /// </summary>
        public void Clear()
        {
            //smazani starych hodnot
            queryString.Clear();
            addRange = 0;
            method = MethodType.Get;
            referer = null;
        }
    }

    ///<summary>
    /// Represents the method that will handle the OSD.PluginBase.WebDownloader.SaveToFileProgressChange event of a OSD.PluginBase.WebDownloader.SaveToFile method.
    ///</summary>
    ///<param name="sender">current WebDownloader object</param>
    ///<param name="args">object of SaveToFileProgressChangedEventArgs</param>
    public delegate void SaveToFileProgressChange(object sender, SaveToFileProgressChangedEventArgs args);

    /// <summary>
    /// Provides data for the OSD.PluginBase.WebDownloader.SaveToFileProgressChange event of a OSD.PluginBase.WebDownloader.SaveToFile method.
    /// </summary>
    public class SaveToFileProgressChangedEventArgs
    {
        private readonly long bytesReceived;
        private readonly long totalBytesToReceive;

        /// <summary>
        /// Main Constructor of SaveToFileProgressChangedEventArgs class
        /// </summary>
        /// <param name="totalBytesToReceive">total bytes which will be received</param>
        /// <param name="bytesReceived">number of bytes received</param>
        public SaveToFileProgressChangedEventArgs(long totalBytesToReceive, long bytesReceived)
        {
            this.totalBytesToReceive = totalBytesToReceive;
            this.bytesReceived = bytesReceived;
        }

        /// <summary>
        /// Gets the total bytes which will be received.
        /// </summary>
        public long TotalBytesToReceive
        {
            get { return totalBytesToReceive; }
        }

        /// <summary>
        /// Gets the number of bytes received.
        /// </summary>
        public long BytesReceived
        {
            get { return bytesReceived; }
        }
    }

    /// <summary>
    /// Describes an exception that occurred when response stream is not available.
    /// </summary>
    public class ResponseStreamNotAvaibleException : HttpException
    {
        /// <summary>
        /// Initializes a new instance of this class.
        /// </summary>
        public ResponseStreamNotAvaibleException() {}
        /// <summary>
        /// Initializes a new instance of this class using a supplied error message.
        /// </summary>
        /// <param name="message">The error message displayed to the client when the exception is thrown.</param>
        public ResponseStreamNotAvaibleException(string message) : base(message) {}
        /// <summary>
        /// Initializes a new instance of this class using an error message and the System.Exception.InnerException property.
        /// </summary>
        /// <param name="message">The error message displayed to the client when the exception is thrown.</param>
        /// <param name="innerException">The System.Exception.InnerException, if any, that threw the current exception.</param>
        public ResponseStreamNotAvaibleException(string message, Exception innerException)
            : base(message, innerException) {}
    }
}