﻿// $Id: Settings.cs 20 2008-03-26 02:51:32Z arcao $
//
// Copyright (C) 2008 Martin "Arcao" Sloup <arcao@arcao.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Globalization;
using System.Xml;

namespace OSD.Program
{
    /// <summary>
    /// Trida s nastavenim v XML souboru
    /// </summary>
    public class Settings : PluginBase.Settings
    {
		/// <summary>
		/// soubor s nastavenim aplikace
		/// </summary>
        public static string default_filename = "config.xml";
        private static Settings settings = null;

        private readonly XmlDocument xmlDoc;
        private string filename = default_filename;

        /// <summary>
        /// Konstruktor tridy se zadanim souboru s nastavenim
        /// </summary>
        /// <param name="filename">XML soubor, odkud bude nastaveni nacteno</param>
        public Settings(string filename)
        {
            xmlDoc = new XmlDocument();
            Load(filename);
        }

        private Settings() : this(default_filename) {}

        ///<summary>
        /// Vraci XMLDocument, kde je ulozen strom s nastavenim
        ///</summary>
        public override XmlDocument XmlDocument
        {
            get { return xmlDoc; }
        }

        /// <summary>
        /// Vrati statickou instanci teto tridy. Zaroven se pokusi nacist nastaveni programu ze souboru zadanem v default_filename. 
        /// </summary>
        /// <returns>instance teto tridy</returns>
        public static Settings getInstance()
        {
            if (settings == null)
                settings = new Settings();
            return settings;
        }

        /// <summary>
        /// Nacte nastaveni programu ze zadaneho xml souboru
        /// </summary>
        /// <param name="filename">XML soubor, odkud bude nastaveni nacteno</param>
        public void Load(string filename)
        {
            try
            {
                xmlDoc.Load(filename);
                this.filename = filename;
            }
            catch {}
        }

        /// <summary>
        /// Nacte nastaveni programu z defaultniho nazvu xml souboru
        /// </summary>
        public void Load()
        {
            Load(filename);
        }

        /// <summary>
        /// Ulozi nastaveni programu do zadaneho XML souboru
        /// </summary>
        /// <param name="filename">XML soubor, kam bude nastaveni ulozeno</param>
        public void Save(string filename)
        {
            if (!xmlDoc.OuterXml.Contains("<?xml"))
                xmlDoc.InsertBefore(xmlDoc.CreateXmlDeclaration("1.0", "utf-8", null), xmlDoc.DocumentElement);
            xmlDoc.Save(filename);
            this.filename = filename;
        }

        ///<summary>
        /// Ulozi nastaveni programu do otevreneho Xml souboru
        ///</summary>
        public override void Save()
        {
            Save(filename);
        }

        /// <summary>
        ///  Vraci stringovou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud nastane chyba, vraci vychozi hodnotu z promene <code>default_value</code>. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kde je ulozena hodnota</param>
        /// <param name="default_value">Vychozi hodnota, pokud node neexistuje, nebo nastala-li chyba</param>
        /// <returns>Hodnota pro zadany node</returns>
        public override string getString(string name, string default_value)
        {
            try
            {
                return xmlDoc.SelectSingleNode(name).InnerText;
            }
            catch
            {
                return default_value;
            }
        }

        /// <summary>
        ///  Vraci celociselnou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud nastane chyba, vraci vychozi hodnotu z promene <code>default_value</code>. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kde je ulozena hodnota</param>
        /// <param name="default_value">Vychozi hodnota, pokud node neexistuje, nebo nastala-li chyba</param>
        /// <returns>Hodnota pro zadany node</returns>
        public override int getInt(string name, int default_value)
        {
            try
            {
                return int.Parse(xmlDoc.SelectSingleNode(name).InnerText);
            }
            catch
            {
                return default_value;
            }
        }

        /// <summary>
        ///  Vraci booleovskou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud nastane chyba, vraci vychozi hodnotu z promene <code>default_value</code>. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kde je ulozena hodnota</param>
        /// <param name="default_value">Vychozi hodnota, pokud node neexistuje, nebo nastala-li chyba</param>
        /// <returns>Hodnota pro zadany node</returns>
        public override bool getBool(string name, bool default_value)
        {
            try
            {
                return bool.Parse(xmlDoc.SelectSingleNode(name).InnerText);
            }
            catch
            {
                return default_value;
            }
        }

        /// <summary>
        ///  Vraci ciselnou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud nastane chyba, vraci vychozi hodnotu z promene <code>default_value</code>. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kde je ulozena hodnota</param>
        /// <param name="default_value">Vychozi hodnota, pokud node neexistuje, nebo nastala-li chyba</param>
        /// <returns>Hodnota pro zadany node</returns>
        public override float getFloat(string name, float default_value)
        {
            try
            {
                return
                    float.Parse(xmlDoc.SelectSingleNode(name).InnerText,
                                CultureInfo.GetCultureInfo("en-US").NumberFormat);
            }
            catch
            {
                return default_value;
            }
        }

        /// <summary>
        ///  Nastavuje stringovou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud zadana XPath cesta neexistuje, vytvori ji. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kam bude ulozena hodnota</param>
        /// <param name="value">Hodnota pro zadany node</param>
        public override void setString(string name, string value)
        {
            try
            {
                createSelectNodes(xmlDoc, name).InnerText = value;
            }
            catch {}
        }

        /// <summary>
        ///  Nastavuje celociselnou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud zadana XPath cesta neexistuje, vytvori ji. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kam bude ulozena hodnota</param>
        /// <param name="value">Hodnota pro zadany node</param>
        public override void setInt(string name, int value)
        {
            try
            {
                createSelectNodes(xmlDoc, name).InnerText = value.ToString();
            }
            catch {}
        }

        /// <summary>
        ///  Nastavuje booleovskou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud zadana XPath cesta neexistuje, vytvori ji. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kam bude ulozena hodnota</param>
        /// <param name="value">Hodnota pro zadany node</param>
        public override void setBool(string name, bool value)
        {
            try
            {
                createSelectNodes(xmlDoc, name).InnerText = value.ToString();
            }
            catch {}
        }

        /// <summary>
        ///  Nastavuje ciselnou hodnotu XPath dotazu na XmlNode zadane do <code>name</code>. Pokud zadana XPath cesta neexistuje, vytvori ji. 
        /// </summary>
        /// <param name="name">XPath dotaz na node, kam bude ulozena hodnota</param>
        /// <param name="value">Hodnota pro zadany node</param>
        public override void setFloat(string name, float value)
        {
            try
            {
                createSelectNodes(xmlDoc, name).InnerText =
                    value.ToString(CultureInfo.GetCultureInfo("en-US").NumberFormat);
            }
            catch {}
        }

        /// <summary>
        ///  Vrati XmlNode zadany formou jednoducheho XPath dotazu <code>path</code>. Pokud XmlNode neexistuje, vytvori ho.
        /// </summary>
        /// <param name="xmlDoc">XmlDokument, na kterem se bude pokladat dotaz</param>
        /// <param name="path">Cesta v Xml</param>
        /// <returns>Node vyhovujici prislusne ceste</returns>
        public override XmlNode createSelectNodes(XmlDocument xmlDoc, string path)
        {
            if (path.Substring(0, 2) == "//")
                path = path.Substring(2);

            if (xmlDoc.SelectNodes(path).Count != 0)
                return xmlDoc.SelectSingleNode(path);

            if (path.IndexOf("/") == -1)
                return xmlDoc.AppendChild(xmlDoc.CreateElement(path));
            else
            {
                String[] path_items = path.Split('/');
                XmlNode node = xmlDoc;
                for (int i = 0; i < path_items.Length; i++)
                {
                    if (node.SelectNodes(path_items[i]).Count != 0)
                        node = node.SelectSingleNode(path_items[i]);
                    else
                        node = node.AppendChild(xmlDoc.CreateElement(path_items[i]));
                }
                return node;
            }
        }

        /// <summary>
        ///  Vrati Node vyhovujici prislusne ceste
        /// </summary>
        /// <param name="path">Cesta v Xml stromu</param>
        /// <returns>Node vyhovujici prislusne ceste</returns>
        public override XmlNode getNode(string path)
        {
            return xmlDoc.SelectSingleNode(path);
        }

        /// <summary>
        ///  Vrati NodeList vyhovujici prislusne ceste
        /// </summary>
        /// <param name="path">Cesta v Xml stromu</param>
        /// <returns>NodeList vyhovujici prislusne ceste</returns>
        public override XmlNodeList getNodes(string path)
        {
            return xmlDoc.SelectNodes(path);
        }
    }
}